from peewee import *
from arikaim.core.container import di

class Users(Model):     
    id = BigAutoField(unique = True, primary_key = True)
    uuid = CharField(unique = True)
    user_name = CharField(unique = True)
    email = CharField(unique = True)
    password = CharField(unique = False)
    status = IntegerField()
    date_login = IntegerField()
    date_created = IntegerField()
    date_deleted = IntegerField()

    class Meta:
        table_name = 'users'
        database = di.get('db').peewee


from arikaim.core.controller import Controller, get
from arikaim.core.container import di

class ServicesList(Controller):

    @get
    async def get(self, request, data):  
        app = di.get('app').app()

        self.field('services',app.services)
        self.message('Service list')
        


from starlette.authentication import AuthenticationBackend, AuthCredentials
from arikaim.core.access.auth_error import AuthError
from arikaim.core.container import di

class MultipleAuthMiddleware(AuthenticationBackend):
    
    def __init__(self, auth_providers = []):
        self._providers = auth_providers

    def add_provider(self, auth_provider):
        if not auth_provider in self._providers:
            self._providers.append(auth_provider)

    async def authenticate(self, conn):      
        for provider in self._providers:
            user = di.get('access').authenticate(conn,provider)
            if user != False:
                return AuthCredentials(["authenticated"]), user

        raise AuthError(status_code = 401)



from starlette.endpoints import HTTPEndpoint
from starlette.types import Receive, Scope, Send

from arikaim.core.api_descriptor import ApiDescriptor
from arikaim.core.response import ApiResponse

class Controller(HTTPEndpoint):
  
    def __init__(self, scope: Scope, receive: Receive, send: Send):
        """Constructor. """
        super().__init__(scope, receive, send)
        # response fields
        self._status = 'ok'
        self._errors = []
        self._code = 200 
        self._fields = {}
        self._descriptor = None

        self.boot()
        
    def boot(self):
        pass

    def response(self):       
        return ApiResponse(self.get_result())

    def message(self, value):
        self._fields['message'] = value

    def field(self, key, value):
        self._fields[key] = value

    def error(self, error):
        self._errors.append(error)

    @property
    def status(self,value):
        self._status = value
    
    @property
    def code(self, value):
        self._code = value

    def get_result(self):  

        if (len(self._errors) > 0):
            self._status = 'error'

        return {
            'result': self._fields,
            'status': self._status,
            'errors': self._errors,
            'code'  : self._code
        }

    @property
    def descriptor(self):
        if self._descriptor == None:
            self._descriptor = ApiDescriptor()

        return self._descriptor
    
    @classmethod
    def init_descriptor(cls, descriptor):               
        pass

# Decorators
def put(func):
    async def wrap(self, *args):
        query_params = args[0].query_params
        path_params = args[0].path_params
        body = await args[0].json()
        data = {**query_params,**path_params, **body}

        await func(self,args[0],data)
        return self.response()
    return wrap

def post(func):
    async def wrap(self, *args):
        query_params = args[0].query_params
        path_params = args[0].path_params
        body = await args[0].json()
        data = {**query_params,**path_params, **body}

        await func(self,args[0],data)
        return self.response()
    return wrap

def get(func):
    async def wrap(self, *args):
        query_params = args[0].query_params
        path_params = args[0].path_params
        data = {**query_params,**path_params}

        await func(self,args[0],data)      
        return self.response()
    return wrap

def delete(func):
    async def wrap(self, *args):
        query_params = args[0].query_params
        path_params = args[0].path_params
        data = {**query_params,**path_params}

        await func(self,args[0],data)
        return self.response()
    return wrap


import click
import os
from rich import print
from arikaim.core.path import Path
from arikaim.core.console.templates .config_file import content

@click.group()
def config():
    pass

@click.command()
def create():
    print('')
    print('Create config file')
    print('')

    file_name = Path.config('config.py')

    if (os.path.exists(file_name) == True):
        print('[red]Config file exist')
        print('')
        return False
    
    with open(file_name,'w') as config_file:  
        config_file.write(content)
        
    print('[green]Done.')
    print('')

config.add_command(create)



from arikaim.core.collection.property import Property

class ApiDescriptor:

    def __init__(self):
        self._params = {}
        self._result = {}
        self._title = ''
        self._description = ''

    @property
    def title(self):
        return self._title
    
    @title.setter
    def title(self, value: str):
        self._title = value

    @property
    def description(self):
        return self._description
    
    @description.setter
    def description(self, value: str):
        self._description = value

    def param(self, name: str):
        self._params[name] = Property(name)
        return self._params[name]
    
    def result(self, name: str):
        self._result[name] = Property(name)
        return self._result[name]
    
    @property
    def params(self):
        return self._params

    @property
    def result_fields(self):
        return self._result



def get_descriptor(cls):
    descriptor = ApiDescriptor()
    cls.init_descriptor(descriptor)
   
    return descriptor




class Job:

    def __init__(self):
        self._schedule_time = None
        self._name = None
        self._interval = None

    def __call__(self):
        self.execute()

    def execute(self):
        pass

    @property
    def interval(self):
        return self._interval


import sys, os, importlib, imp
from string import Template
from datetime import datetime
from arikaim.core.path import Path

def inlcude_service_lib(service_name: str, module_name: str):
    path = os.path.join(Path.services(service_name),'lib',module_name + '.py')
    return load_module(module_name,path)

def load_service_config(service_name: str):
    return load_module('config',os.path.join(Path.services(service_name),'config.py'))

def call(object, method_name: str, **params):
    method = getattr(object,method_name,None)
    if not method:
        return False

    return method(**params)

def time_now() -> int:
    return datetime.now().timestamp()

def load_class(path, module_name, class_name):
    module_path = os.path.join(path)

    sys.path.append(module_path)
    module = importlib.import_module(module_name) 
   
    return getattr(module,class_name)

def load_module_vars(path, module_name):
    module_path = os.path.join(path)
    sys.path.append(module_path)

    return importlib.import_module(module_name, package = module_name)
   
def load_module(name, path):
    try:
        return imp.load_source(name,path)
    except BaseException as e:
        print(format(e))       
        return None
    
def parse_text(text, vars):
    template = Template(text)
    return template.substitute(vars)

def singleton(class_):
    class class_w(class_):
        _instance = None
        def __new__(class2, *args, **kwargs):
            if class_w._instance is None:
                class_w._instance = super(class_w, class2).__new__(class2, *args, **kwargs)
                class_w._instance._sealed = False
            return class_w._instance
        def __init__(self, *args, **kwargs):
            if self._sealed:
                return
            super(class_w, self).__init__(*args, **kwargs)
            self._sealed = True
    class_w.__name__ = class_.__name__
    return class_w

def php_unserialize(text, delimiter = ';', var_delimiter = '|', value_delimiter = ':'):
    vars = {}
    text = text.replace('{','')
    text = text.replace('}','')

    items = text.split(delimiter)
    for item in items:
        sub_item = item.split(var_delimiter)
        key = sub_item[0]
        if key != '':          
            if len(sub_item) >= 2:
                value = sub_item[1].split(value_delimiter)
            else:
                value = ''
            if len(value) > 0:
                vars[key] = value[1]

    return vars


import logging

uvicorn_logger = logging.getLogger("uvicorn")
uvicorn_logger.propagate = False

logging.basicConfig(
    format = '{levelname}: {message}', 
    style = '{', 
    level = logging.INFO
)

logger = logging.getLogger(__name__)




class Property:

    TYPE_NAMES = [
        'text',
        'number',
        'custom',
        'boolean',
        'list',
        'class',
        'password',
        'url',
        'text-area',
        'group',
        'oauth',
        'language-dropdown',
        'image'
    ]

    def __init__(self, name: str, data = None):
        self._name = name
        self._title = ''
        self._description = ''
        self._type = 'text'
        self._readonly = False
        self._required = False
        self._value = None
        
        if type(data) is dict:
            for key in data:
                setattr(self,key,data[key])

    def to_json(self):
        return {
            'name': self._name,
            'description': self._description,
            'type': self._type,
            'readonly': self._readonly,
            'value': self._value,
            'required': self._required,
            'title': self._title
        }
       
    def value(self, value: str):
        self._value = value
        return self
    
    def title(self, value: str):
        self._title = value
        return self
    
    def description(self, value: str):
        self._description = value

    def required(self, value: bool):
        self._required = value
        return self
    
    def readonly(self, value: bool):
        self._readonly = value
        return self
    
    def type(self, type_name: str):
        self._type = self.get_type_id(type_name)
        return self
    
    def get_type_id(self, type_name: str):
        try: 
            return self.TYPE_NAMES.index(type_name) 
        except ValueError: 
            return None       


content = """# Arikaim services server config file

# app settings
settings = {
    'port': 5000,
    'host': '127.0.0.1'
}

# db settings
db = {
    'type': 'mysql',
    'host': '127.0.0.1',
    'username': 'user name',
    'password': 'password',
    'database': 'db name',
    'synchronize': True,
    'logging': False,
    'entities': []
}

# global middlewares
middleware = {
}
"""


from peewee import *
from arikaim.core.container import di

class AccessTokens(Model):
    id = BigAutoField(unique = True, primary_key = True)
    uuid = CharField(unique = True)
    token = CharField(unique = True)
    status = IntegerField()
    type = IntegerField()
    user_id = IntegerField()
    date_created = IntegerField()
    date_expired = IntegerField()

    class Meta:
        table_name = 'access_tokens'
        database = di.get('db').peewee
    


import click
from rich import print
from arikaim.core.server import arikaim_server
from arikaim.core.app import app
from arikaim.core.console.config import config
from arikaim.core.console.install import install
from arikaim.core.console.packages import packages
from arikaim.core.utils import call
from arikaim.core.logger import logger

@click.group(invoke_without_command = True)
@click.pass_context
def main(ctx):
    print("")
    print("[blue]Arikaim Services Server")
    print("[green]version [white]" + arikaim_server.version)
    print("")

    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())

@click.command()
@click.argument('mode', required = False)
@click.option('--path', required = False)
def run(mode = None, path = None):
    # run server
    if mode == 'dev':
        mode = True
 
    arikaim_server.run(mode,path)

@click.command()
@click.argument('service-name')
@click.argument('module-name')
def cli(service_name: str, module_name: str):
    logger.info('Service: ' + service_name + ' load console commands ...')
    module = app.load_console_commands(service_name,module_name)
    call(module,'main')
  
main.add_command(run)
main.add_command(cli)
main.add_command(packages)
main.add_command(config)
main.add_command(install)


import types
from inspect import isclass
from arikaim.core.utils import singleton

@singleton
class Container:
 
    def __init__(self):
        self._services = {} # Services registry

    def register(self, name: str):
        def wrap(fn):  
            self.add(name,fn)
        return wrap

    def add(self, name: str, item):
        if self.has(name):
            raise False
        if isclass(item):
            def class_wrap(self):
                return item()

            self._services[name] = class_wrap
        else:
            self._services[name] = item
  
    def get(self, name: str):
        if not self.has(name):
            raise RuntimeError("Service '{}' not found." . format(name))
        item = self._services[name]
        
        if type(item) in [types.LambdaType, types.FunctionType]:
            item = item()
        self._services.pop(name)  
        self._services[name] = item

        return item

    def has(self, name: str) -> bool:
        return name in self._services.keys()


di = Container()


from arikaim.core.controller import Controller, get
from arikaim.core.container import di
from arikaim.core.api_descriptor import *

class ServiceRoute(Controller):

    @get
    async def get(self, request, data):  
        app = di.get('app').app()
        service = app.get_service(data['name'])
        path = data['path'].lstrip('/')
        result = {} 
        for route in service.routes.routes:
            if route.path.lstrip('/') == path:
                descriptor = get_descriptor(route.endpoint)
                result = {
                    'title': descriptor.title,
                    'description': descriptor.description,
                    'full_path': service.mount + route.path,     
                    'path': route.path,
                    'endpoint': route.name,
                    'methods': list(route.methods),
                    'params': descriptor.params,
                    'result': descriptor.result_fields          
                }
                break

        self.field('name',data['name'])
        self.field('path',data['path'])
        self.field('server_url',app.server_url)
        self.field('route',result)
        self.message('Service route')
    


import uvicorn
from peewee import *

from arikaim.core.logger import logger
from arikaim.core.utils import *
from arikaim.core.app import app

class ArikaimServer:
    _instance = None

    def __init__(self,config):
        self._version = '0.5.7'
        self._config = config
        
    def run(self, reload = False, path = None):
        if path is not None:          
            Path.base_path = path
            logger.info('Project path ' + Path.base())
            
        if reload == True:
            logger.info('Dev mode (reload)')

        uvicorn.run(
            'arikaim.core.app:app.boot', 
            reload = reload,
            factory = True,
            host = self._config.settings['host'], 
            port = self._config.settings['port'], 
            log_level = self._config.settings.get('log_level','info')
        )
       
    @property
    def server_url(self):
        return 'http://' + self._config.settings['host'] + ':' + str(self._config.settings['port'])
    
    @property 
    def version(self):
        return self._version

    @classmethod
    def instance(cls):
        if cls._instance is None:           
            cls._instance = ArikaimServer(app.config)
          
        return cls._instance
    
arikaim_server = ArikaimServer.instance()


import os

class Path:
    base_path = ''

    @staticmethod
    def base():
        if Path.base_path is None:
            Path.base_path = os.getcwd()
            
        return Path.base_path

    @staticmethod
    def arikaim():
        return os.path.join(Path.base(),'arikaim')
    
    @staticmethod
    def model_path(service_name):
        if not service_name:
            return os.path.join(Path.base(),'db','models')
        else:
            return os.path.join(Path.services(service_name),'models')

    @staticmethod
    def job_path(service_name):
        if not service_name:
            return os.path.join(Path.base(),'queue','jobs')
        else:
            return os.path.join(Path.services(service_name),'jobs')

    @staticmethod
    def console_path(service_name):
        return os.path.join(Path.services(service_name),'console')

    @staticmethod
    def storage_path(path: str = ''):       
        return os.path.join(Path.base(),'arikaim','storage',path.strip('/'))
        
    @staticmethod
    def services(name = ''):
        services_path = os.path.join(Path.base(),'arikaim','services')
        if name != "":
            return os.path.join(services_path,name)
        return services_path

    @staticmethod
    def config(file_name = None):
        path = os.path.join(Path.base(),'arikaim','config')
        if (file_name != None): 
            path = os.path.join(path,file_name) 
        return path
        


from arikaim.core.db.models.users import Users
from arikaim.core.db.models.access_tokens import AccessTokens


class TokenAuthProvider:
    
    def authenticate(self, credentails):

        if "Authorization" not in credentails.headers:
            return False

        try:
            token = credentails.headers["Authorization"]
            access_tokens = AccessTokens.select().where(AccessTokens.token == token).get()
            
            return Users.get(Users.id == access_tokens.user_id)
        except (Users.DoesNotExist, AccessTokens.DoesNotExist):
            return False


from peewee import *
from arikaim.core.container import di

class Permissions(Model):
    id = BigAutoField(unique = True, primary_key = True)
    uuid = CharField(unique = True)
    name = CharField(unique = True)
    slug = CharField(unique = True)
    title = CharField()
    extension_name = CharField()
    description = CharField()
    deny = IntegerField()

    class Meta:
        table_name = 'permissions'
        database = di.get('db').peewee


from peewee import *
from arikaim.core.container import di

class Jobs(Model):     
    id = BigAutoField(unique = True, primary_key = True)
    uuid = CharField(unique = True)
    status = IntegerField()
    name = CharField(unique = True)
    handler_class = CharField()
    recuring_interval = CharField()
    extension_name = CharField()
    priority = IntegerField()
    executed = IntegerField()
    type = CharField()
    service_name = CharField()
    date_created = IntegerField()
    schedule_time = IntegerField()
    date_executed = IntegerField()
    config = TextField()
    queue = CharField()

    @staticmethod
    def get_jobs_due():
        try:
            model = Jobs.select().where(Jobs.status == 1).get()

            return model
        except (Jobs.DoesNotExist):                 
            return False

    class Meta:
        table_name = 'jobs'
        database = di.get('db').peewee


from peewee import *
from arikaim.core.container import di

class PermissionRelations(Model):
    id = BigAutoField(unique = True, primary_key = True)
    uuid = CharField(unique = True)
    read = IntegerField()
    write = IntegerField()
    delete = IntegerField()
    execute = IntegerField()
    permission_id = IntegerField()
    relation_id = IntegerField()
    relation_type = CharField()

    class Meta:
        table_name = 'permission_relations'
        database = di.get('db').peewee


import json
import os
import pip
from arikaim.core.path import Path
from arikaim.core.logger import logger

def load_package_descriptor(service_name: str):
    package_file = os.path.join(Path.services(service_name),'arikaim-package.json')
    file = open(package_file,'r')
   
    descriptor = json.loads(file.read())
    file.close()

    return descriptor

   
def install_servide_packages(descriptor):   
    packages = descriptor['require']['packages'];

    for package in packages:        
        logger.info('Install Package ' + package)
        
        if hasattr(pip, 'main'):
            pip.main(['install', package])
        else:
            pip._internal.main(['install', package])


def install_packages(app):
    
    for service_name in app.services:
        logger.info('Service ' + service_name)
        descriptor = load_package_descriptor(service_name)

        install_servide_packages(descriptor)
    


from os.path import exists
import subprocess
import json
import arikaim.core.globals as globals

from arikaim.core.db.models.users import Users
from arikaim.core.db.models.access_tokens import AccessTokens
from arikaim.core.utils import php_unserialize
from arikaim.core.logger import logger

class PhpSessionAuthProvider:
    
    def __init__(self):
        self._php_session_path = None
        self.resolve_session_path()

    def authenticate(self, credentails):
        session_id = credentails.cookies.get('PHPSESSID')
        if not session_id:
            return False
        try:
            session_data = self.read_php_session(session_id)
            if 'auth.id' not in session_data:
                return False
  
            return Users.get(Users.id == session_data['auth.id'])

        except (Users.DoesNotExist, AccessTokens.DoesNotExist):
            return False

    def read_php_session(self, session_id):
        file_name = self._php_session_path + '/sess_' + session_id

        if exists(file_name) == False:
            return {}
      
        logger.info('Read php session file: ' + file_name)

        file = open(file_name,'r')
        content = file.read()
        file.close()
        
        return php_unserialize(content)
    
    def resolve_session_path(self):
        response = subprocess.Popen(
            'php cli session:info --output json',
            shell = True,
            cwd = globals.ARIKAIM_PATH,
            stdout = subprocess.PIPE          
        )

        json_text = response.stdout.read()
        data = json.loads(json_text)
        self._php_session_path = data['save_path']


from arikaim.core.controller import Controller, get
from arikaim.core.container import di
from arikaim.core.api_descriptor import *

class ServiceRoutes(Controller):

    @get
    async def get(self, request, data):  
        app = di.get('app').app()
        service = app.get_service(data['name'])
      
        routes = []
        for route in service.routes.routes:
            descriptor = get_descriptor(route.endpoint)

            routes.append({
                'title': descriptor.title,
                'description': descriptor.description,
                'path': route.path,    
                'full_path': service.mount + route.path,
                'endpoint': route.name,
                'methods': list(route.methods),
                'params': descriptor.params,
                'result': descriptor.result_fields          
            })

        self.field('name',data['name'])
        self.field('server_url',app.server_url)
        self.field('routes',routes)
        self.message('Service routes')
       
    


import json
import typing

from starlette.responses import JSONResponse

class ApiResponse(JSONResponse):
    media_type = "application/json"

    def render(self, content: typing.Any) -> bytes:
        return json.dumps(content,
            ensure_ascii = False,
            allow_nan = False,
            indent = None,
            default = lambda obj : obj.to_json(),
            separators = (",", ":"),
        ).encode("utf-8")
        


from starlette.exceptions import HTTPException
from starlette.requests import Request
from arikaim.core.response import ApiResponse

async def unauthorized_error(request: Request, exc: HTTPException): 
    return ApiResponse({
            'result': '',
            'status': 401,
            'errors': ['Access Denied'],
            'code'  : 401
    })  

async def not_found_error(request: Request, exc: HTTPException): 
    return ApiResponse({
            'result': '',
            'status': 404,
            'errors': ['Not found'],
            'code'  : 404
    })  

error_handlers = {
    401: unauthorized_error,
    404: not_found_error,
}


from arikaim.core.utils import *
from starlette.routing import Route, Mount, WebSocketRoute
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware

class Service:
    
    def __init__(self, name, mount_path = None):
        self._routes = []      
        self._middlewares = []     
        self._name = name
        self._mounth_path = mount_path
        
    def boot(self):
        pass

    def init_routes(self):
        pass

    def init_container(self):
        pass

    def add_route(self, method, path, endpoint):
        self._routes.append(Route(path,endpoint,methods = method))           

    def add_websocket_route(self, path, endpoint):       
        self._routes.append(WebSocketRoute(path,endpoint = endpoint))

    def add_middleware(self, middleware_class):
        self._middlewares.append(Middleware(middleware_class))     

    def add_auth_middleware(self, backend):
        self._middlewares.append(Middleware(AuthenticationMiddleware, backend = backend))  

    @property 
    def mount_path(self, path):
        self._mounth_path = path

    @property
    def mount(self):
        if self._mounth_path == None:
            return self._name
        else:
            return self._mounth_path
        
    @property
    def routes(self):
        if self._mounth_path == None:
            self._mounth_path = self._name
            
        if len(self._routes) == 0:
            return None
        else:
            return Mount(
                "/" + self._mounth_path, 
                routes = self._routes, 
                name = self._name, 
                middleware = self._middlewares
            )


from arikaim.core.container import di
from arikaim.core.service import Service
from arikaim.core.admin.controllers.services_list import ServicesList
from arikaim.core.admin.controllers.service import ServiceDescriptor
from arikaim.core.admin.controllers.service_routes import ServiceRoutes
from arikaim.core.admin.controllers.service_route import ServiceRoute
from arikaim.core.admin.controllers.info import Info

class AdminService(Service):

    def init_routes(self):

        auth = di.get('access').middleware('multiple',['token','php_session'])
        self.add_auth_middleware(auth)

        self.add_route(["GET"],"/info",Info)
        self.add_route(["GET"],"/services",ServicesList)
        self.add_route(["GET"],"/service/{name}",ServiceDescriptor)
        self.add_route(["GET"],"/service/routes/{name}",ServiceRoutes)
        self.add_route(["GET"],"/service/route/{name}/{path}",ServiceRoute)


from arikaim.core.utils import *
from arikaim.core.db.db import load_model_class
from arikaim.core.access.auth_error import AuthError

class Access:
    
    DEFAULT_AUTH_PROVIDER = 'token'

    def __init__(self):
        self._auth_middleware_classes = {
            'multiple': 'MultipleAuthMiddleware'
        }
        self._auth_provider_classes = {
            'token': 'TokenAuthProvider',
            'php_session': 'PhpSessionAuthProvider'
        }

    def authenticate(self, credentails, auth_name = None):
        if auth_name == None:
            auth_name = Access.DEFAULT_AUTH_PROVIDER

        provider = self.provider(auth_name)
                
        return provider.authenticate(credentails)


    def middleware(self, auth_name, auth_providers):
        if auth_name in self._auth_middleware_classes:
            middleware_class = Access.load_moddleware_class(auth_name,self._auth_middleware_classes[auth_name])
            return middleware_class(auth_providers)
    
        else:
            return None


    def provider(self, auth_name):
        if auth_name in self._auth_provider_classes:
            provider_class = Access.load_provider_class(auth_name,self._auth_provider_classes[auth_name])
            return provider_class()

        else:
            return None


    def has_permission(self, name, user_id):
        Permissions = load_model_class('Permissions','permissions')
        PermissionRelations = load_model_class('PermissionRelations','permission_relations')
    
        try:
            permission = Permissions.select().where(Permissions.name == name).get()
            relation = PermissionRelations.get(
                PermissionRelations.relation_id == user_id and 
                PermissionRelations.permission_id == permission.id
            )
            return relation.relation_id == user_id

        except (Permissions.DoesNotExist, PermissionRelations.DoesNotExist):
            return False

    def is_admin_user(self, user_id):
        return self.has_permission('ControlPanel',user_id)

    def require_admin_user(self, user_id):       
        if self.is_admin_user(user_id) == False:          
            raise AuthError(status_code = 401)            


    @staticmethod
    def load_moddleware_class(module_name, class_name):
        module = importlib.import_module('arikaim.core.access.middleware.' + module_name,class_name)
        return getattr(module,class_name)


    @staticmethod
    def load_provider_class(module_name, class_name):
        module = importlib.import_module('arikaim.core.access.providers.' + module_name,class_name)
        return getattr(module,class_name)


import click
from arikaim.core.app import app
from arikaim.core.packages import *

@click.group()
def packages():
    pass

@click.command()
def install():
    # boot server
    install_packages(app)

packages.add_command(install)


from peewee import *
from arikaim.core.path import Path
import imp,os,importlib;

class Db: 
    def __init__(self, config):
        self._config = config
        self._peewee = None

    def connect(self):     
        self._peewee = MySQLDatabase(
            self._config['database'],
            user = self._config['username'],
            password = self._config['password'],
            host = self._config['host']
        )

    def close(self):
        if not self._peewee:
            return False
        self._peewee.close()
        
        return True

    def bind(self, models):
        self._peewee.bind(models)

    @property
    def peewee(self):
        return self._peewee


def load_model_class(model_class, module_name, service_name = None):
    if not service_name:        
        module = importlib.import_module('arikaim.core.db.models.' + module_name,model_class)
    else:
        path = os.path.join(Path.model_path(service_name),module_name)      
        module = imp.load_source(module_name,path + '.py')

    return getattr(module,model_class)


from distutils.core import setup
from setuptools import find_packages
import os

# Optional project description in README.md:
current_directory = os.path.dirname(os.path.abspath(__file__))

try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''

setup(

scripts = [
    'arikaim-server',
],
# Project name: 
name = 'arikaim',

# Packages to include in the distribution: 
packages = [
    'arikaim.core',
    'arikaim.core.admin',
    'arikaim.core.admin.controllers',
    'arikaim.core.collection',
    'arikaim.core.access',
    'arikaim.core.access.middleware',
    'arikaim.core.access.providers',
    'arikaim.core.console',
    'arikaim.core.console.templates',
    'arikaim.core.db',
    'arikaim.core.queue',
    'arikaim.core.db.models'
],
   
# Project version number:
version = '0.5.7',

# List a license for the project, eg. MIT License
license = 'MIT License',

# Short description of your library: 
description = 'Arikaim services server',

# Long description of your library: 
long_description = long_description,
long_description_content_type = 'text/markdown',

# Your name: 
author = 'Intersoft Ltd',

# Your email address:
author_email = 'info@arikaim.com',

# Link to your github repository or website: 
url = 'https://github.com/arikaim/arikaim-services-server.git',

# Download Link from where the project can be downloaded from:
download_url = '',

# List of keywords: 
keywords=['arikaim'],

# List project dependencies: 
install_requires=[   
    'uvicorn',
    'starlette==0.37.1',
    'pymysql',
    'peewee',
    'click',
    'rq',
    'rq-scheduler',
    'rich',
    'requests',
    'pydantic',
    'python-multipart',
    'pymitter'   
],

# https://pypi.org/classifiers/ 
classifiers=[]
)



import os, imp, datetime
from redis import Redis
from rq import Queue as RqQueue
from rq_scheduler import Scheduler
from rq import Worker
from arikaim.core.db.db import load_model_class
from arikaim.core.logger import logger
from arikaim.core.queue.job import Job
from arikaim.core.path import Path

class Queue:

    def __init__(self):
        self._redis = Redis()
        self._queue = RqQueue(connection = self._redis)
        self._worker = Worker([self._queue], connection = self._redis, name = 'jobs')
        self._scheduler = Scheduler(connection = self._redis) 

    def boot(self):
        logger.info('Init qeueue')
        self.remove_jobs()

        Jobs = load_model_class('Jobs','jobs')
        jobs = Jobs.get_jobs_due()
        if not jobs:
            return
        
        for job in jobs:
            pass

    def run(self):
        logger.info('Run qeueue worker')
        self._worker.work()

    def add_job(self, job = None):
        if isinstance(job,Job) == True:        
            if not job.interval:  
                self._queue.enqueue(job)
            else:
                # recurring job job
                self._scheduler.schedule(
                    scheduled_time = datetime.utcnow(), 
                    func = job,                                                  
                    interval = job.interval, # in seconds
                    repeat = None
                )
        pass

    def get_jobs_count(self):
        return len(self._queue)

    def remove_jobs(self):
        self._queue.empty()

    @property
    def jobs(self):
        return self._queue.jobs

    def create_job(self, job_class_name: str, module_name: str, service_name: str):
        path = os.path.join(Path.job_path(service_name),module_name)      
        module = imp.load_source(module_name,path + '.py')

        job_class = getattr(module,job_class_name)
        if not job_class:
            return None
        
        return job_class()


from arikaim.core.controller import Controller, get
from arikaim.core.container import di
from arikaim.core.packages import load_package_descriptor

class Info(Controller):

    @get
    async def get(self, request, data):  
        app = di.get('app').app()

        self.field('host',app.config.settings['host'])
        self.field('port',app.config.settings['port'])
        self.field('version',app.version)
    


import traceback
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from pymitter import EventEmitter

from arikaim.core.utils import *
from arikaim.core.db.db import *
from arikaim.core.errors import error_handlers
from arikaim.core.logger import logger
from arikaim.core.container import di
from arikaim.core.packages import load_package_descriptor
from arikaim.core.admin.admin import AdminService
from arikaim.core.access.access import Access

class ArikaimApp:

    def __init__(self):        
        self._services = []
        self._services_instance = {}
        self._starlette = None
        self._config = None
        self._routes = []
        self.system_init()

    def load_service_console_commands(self, service_name: str, module_name = None):
        if not module_name:
            module_name = 'console'

        console_file = os.path.join(Path.console_path(service_name),module_name)
        
        if os.path.isfile(console_file + '.py') == True: 
            # append service path  
            sys.path.append(Path.services(service_name))            
            return load_module('console',console_file + '.py')                              
        else:
            return False
        
    def load_console_commands(self, service_name, module_name = None):
        # load console module
        console = self.load_service_console_commands(service_name,'console')
        if console != False:
            call(console,'init')

        return self.load_service_console_commands(service_name, module_name)

    @property 
    def services(self):
        return self._services
    
    def get_service(self, name:str):
        if name in self._services_instance:
            return self._services_instance[name]
        else:
            return None
     
    def system_init(self):
        # create event emiter 
        events = EventEmitter(wildcard = True)
        di.add('events',events)

        # load config
        self._config = load_module('config',os.path.join(Path.config(),'config.py'))
        if self._config is None:
            logger.error('Config file not found!')
            return False

        # connect to db 
        db = Db(self._config.db)
        db.connect()
        # add to container
        di.add('db',db)
        
        #access
        access = Access()
        di.add('access',access)
        # add app
        di.add('app',self)

    def boot(self):
        logger.info('Boot')
  
        # scan services
        for file in os.scandir(Path.services()):
            if file.is_dir() and not file.name.startswith('.'):             
                self._services.append(file.name)
                # add service path to system paths
                sys.path.append(Path.services(file.name))
        
        # load admin routes
        self.load_admin_routes()
        # load services routes
        self.boot_services()

        self._starlette = Starlette(
            debug = True, 
            routes = self._routes, 
            on_shutdown = [self.on_shutdown],
            middleware = [
                Middleware(CORSMiddleware, allow_origins = ['*'])
            ], 
            exception_handlers = error_handlers
        )

        return self.starlette

    async def on_shutdown(self):
        logger.info('Shutdown server')

        di.get('db').close()
        logger.info('Db connection closed')

    def load_admin_routes(self):
        logger.info('Load admin routes')

        admin = AdminService('admin')
        admin.init_routes()

        self._routes.append(admin.routes)

    def boot_services(self, load_routes: bool = True, init_container: bool = True):
      
        for service_name in self._services:            
            package = load_package_descriptor(service_name)
            package.setdefault('language','nodejs')
            package.setdefault('disabled',False)
            
            if package['language'] != 'python':
                # load only pyhton services
                continue
          
            if package['disabled'] == True:
                # skip disabled service boot
                logger.info('service: ' + service_name + ' disabled ')
                continue

            logger.info('Boot service: ' + service_name)
            service_class = load_class(Path.services(service_name),service_name,service_name.capitalize())
            self._services_instance[service_name] = service_class(service_name) 
            
            if load_routes == True:
                self._services_instance[service_name].init_routes()
            if init_container == True:
                self._services_instance[service_name].init_container()

            # add service routes
            if self._services_instance[service_name].routes != None:
                self._routes.append(self._services_instance[service_name].routes)

    @property
    def starlette(self):
        return self._starlette

    @property
    def config(self):
        return self._config


app = ArikaimApp()



import click
import os
from rich import print
from arikaim.core.path import Path
from arikaim.core.container import di
from arikaim.core.db.db import load_model_class

@click.command()
def install():
    print('')
    print('Install ')
    print('')
    print('Create folders ')

    folders = [
        'config',
        'storage',
        'services'
    ]

    for item in folders:
        path = os.path.join(Path.arikaim(),item)

        if (os.path.exists(path) == False):
            os.mkdir(path) 
            print('Create path [green]' + path)

    print('[green]Done.')
    print('')

    print('Create db tables ')
    Users = load_model_class('Users','users')

    di.get('db').peewee.create_table(Users)



from arikaim.core.controller import Controller, get
from arikaim.core.container import di
from arikaim.core.packages import load_package_descriptor

class ServiceDescriptor(Controller):

    @get
    async def get(self, request, data):  
        app = di.get('app').app()

        descriptor = load_package_descriptor(data['name'])

        self.field('name',data['name'])
        self.field('descriptor',descriptor)
    